@@ -213,7 +213,7 @@ class GP():
         method = "global",
         optimization_dict = None,
         pop_size = 20,
-        tolerance = 0.1,
+        tolerance = 0.0001,
         max_iter = 120,
         local_optimizer = "L-BFGS-B",
         global_optimizer = "genetic",
@@ -330,7 +330,7 @@ class GP():
             print("Latest hyperparameters: ", self.hyperparameters)
         except:
             print("Async Hyper-parameter update not successful in fvGP. I am keeping the old ones.")
-            print("That probbaly means you are not optimizing them asynchronously")
+            print("That probably means you are not optimizing them asynchronously")
             print("hyperparameters: ", self.hyperparameters)
         return self.hyperparameters
     ##################################################################################
@@ -503,9 +503,11 @@ class GP():
         b = np.ascontiguousarray(b, dtype=np.float64)
         #dL_dH = self.numba_dL_dH(y, a, b, len(hyperparameters))
         dL_dH = np.empty((len(hyperparameters)))
+        dm_dh = self.dm_dh(hyperparameters)
         for i in range(len(hyperparameters)):
-            dL_dH[i] = 0.5 * ((y.T @ a[i] @ b) - (np.trace(a[i])))
-        return -dL_dH
+            dL_dH[i] = 0.5*(-2.0 * b.T @ dm_dh[i] - y.T @ a[i] @ b) + 0.5 * np.trace(a[i])
+        return dL_dH
+
     ##################################################################################
     @staticmethod
     @nb.njit
@@ -529,6 +531,7 @@ class GP():
         output:
             hessian of the negative marginal log-likelihood (matrix)
         """
+        raise Exception("Hessian not correct, please use the gradient to approximate the Hessian")
         mean = self.mean_function(self,self.data_x,hyperparameters)
         x,K = self._compute_covariance_value_product(hyperparameters,self.data_y, self.variances, mean)
         y = self.data_y - mean
@@ -544,7 +547,11 @@ class GP():
         ss = np.ascontiguousarray(ss, dtype=np.float64)
         #d2L_dH2 = self.numba_d2L_dH2(x, y, s, ss)
         len_hyperparameters = s.shape[0]
-        d2L_dH2 = np.empty((len_hyperparameters,len_hyperparameters))
+        d2L_dH2 =  np.empty((len_hyperparameters,len_hyperparameters))
+        d2L_dm2 =  np.empty((len_hyperparameters,len_hyperparameters))
+        d2L_dmdh = np.empty((len_hyperparameters,len_hyperparameters))
+        d2m_dh2 = self.d2m_dh2(hyperparameters)
+        m1 = self.dm_dh(hyperparameters)
         for i in range(len_hyperparameters):
             x1 = s[i]
             for j in range(i+1):
@@ -552,7 +559,10 @@ class GP():
                 x3 = ss[i,j]
                 f = 0.5 * ((y.T @ (-x2 @ x1 @ x - x1 @ x2 @ x + x3 @ x)) - np.trace(-x2 @ x1 + x3))
                 d2L_dH2[i,j] = d2L_dH2[j,i] = f
-        return -d2L_dH2
+                d2L_dm2[i,j] = d2L_dm2[j,i] = (m1[i,:].T @ np.linalg.inv(K[0]) @ m1[j,:]) - (x.T @ d2m_dh2[i,j])
+                d2L_dmdh[i,j] = d2L_dmdh[j,i] = (x @ x2 @ m1[i,:])
+
+        return -d2L_dH2 - d2L_dm2 + d2L_dmdh
     ##################################################################################
     ##################################################################################
     ##################################################################################
@@ -639,14 +649,14 @@ class GP():
             A = torch.from_numpy(A)
             b = torch.from_numpy(b)
             try:
-                x, lu = torch.solve(b,A)
-                #x = np.linalg.solve(A,b)
+                x = torch.linalg.solve(A,b)
                 return x.numpy()
             except Exception as e:
                 try:
                     print("fvGP: except statement invoked: torch.solve() on cpu did not work")
                     print("reason: ", str(e))
-                    x, qr = torch.lstsq(b,A)
+                    #x, qr = torch.lstsq(b,A)
+                    x, qr = torch.linalg.lstsq(A,b)
                 except Exception as e:
                     print("fvGP: except statement 2 invoked: torch.solve() and torch.lstsq() on cpu did not work")
                     print("falling back to numpy.lstsq()")
@@ -658,12 +668,13 @@ class GP():
             A = torch.from_numpy(A).cuda()
             b = torch.from_numpy(b).cuda()
             try:
-                x, lu = torch.solve(b,A)
+                x = torch.linalg.solve(A, b)
             except Exception as e:
                 print("fvGP: except statement invoked: torch.solve() on gpu did not work")
                 print("reason: ", str(e))
                 try:
-                    x, qr = torch.lstsq(b,A)
+                    #x, qr = torch.lstsq(b,A)
+                    x = torch.linalg.lstsq(A,b)
                 except Exception as e:
                     print("fvGP: except statement 2 invoked: torch.solve() and torch.lstsq() on gpu did not work")
                     print("falling back to numpy.lstsq()")
@@ -680,7 +691,7 @@ class GP():
                 cur_device = torch.device("cuda:"+str(i))
                 tmp_A = torch.from_numpy(tmp_A).cuda(cur_device)
                 tmp_b = torch.from_numpy(tmp_b).cuda(cur_device)
-                results.append(torch.solve(tmp_b,tmp_A)[0])
+                results.append(torch.linalg.solve(tmp_A,tmp_b)[0])
             total = results[0].cpu().numpy()
             for i in range(1,len(results)):
                 total = np.append(total, results[i].cpu().numpy(), 0)
@@ -1451,6 +1462,48 @@ class GP():
             for j in range(i+1):
                 hessian[i,j] = hessian[j,i] = self.d2_gp_kernel_dh2(points1, points2, i,j, hyperparameters)
         return hessian
+
+    def dm_dh(self,hps):
+        gr = np.empty((len(hps),len(self.data_x)))
+        for i in range(len(hps)):
+            temp_hps1 = np.array(hps)
+            temp_hps1[i] = temp_hps1[i] + 1e-6
+            temp_hps2 = np.array(hps)
+            temp_hps2[i] = temp_hps2[i] - 1e-6
+            a = self.mean_function(self,self.data_x,temp_hps1)
+            b = self.mean_function(self,self.data_x,temp_hps2)
+            gr[i] = (a-b)/2e-6
+        return gr
+    ##########################
+    def d2m_dh2(self,hps):
+        hess = np.empty((len(hps),len(hps),len(self.data_x)))
+        e = 1e-4
+        for i in range(len(hps)):
+            for j in range(i+1):
+                temp_hps1 = np.array(hps)
+                temp_hps2 = np.array(hps)
+                temp_hps3 = np.array(hps)
+                temp_hps4 = np.array(hps)
+                temp_hps1[i] = temp_hps1[i] + e
+                temp_hps1[j] = temp_hps1[j] + e
+
+                temp_hps2[i] = temp_hps2[i] - e
+                temp_hps2[j] = temp_hps2[j] - e
+
+                temp_hps3[i] = temp_hps3[i] + e
+                temp_hps3[j] = temp_hps3[j] - e
+
+                temp_hps4[i] = temp_hps4[i] - e
+                temp_hps4[j] = temp_hps4[j] + e
+
+
+                a = self.mean_function(self,self.data_x,temp_hps1)
+                b = self.mean_function(self,self.data_x,temp_hps2)
+                c = self.mean_function(self,self.data_x,temp_hps3)
+                d = self.mean_function(self,self.data_x,temp_hps4)
+                hess[i,j] = hess[j,i] = (a - c - d + b)/(4.*e*e)
+        return hess
+
     ################################################################
     def _normalize_y_data(self):
         mini = np.min(self.data_y)
